/*
 * Copyright 2024 Luca Fulchir <luca.fulchir@runesauth.com>
 *
 * Licensed under the Apache License, Version 2.0 with LLVM exception (the
 * "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License and of the exception at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     https://spdx.org/licenses/LLVM-exception.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

use ::proc_macro::TokenStream;
use ::quote::quote;
use ::syn::{parse_macro_input, DeriveInput, Fields, ItemImpl, ItemStruct};

pub(crate) fn repo_impl(
    _attrs: TokenStream,
    input: TokenStream,
) -> TokenStream {
    let local = parse_macro_input!(input as ItemImpl);

    let reponame = &local.self_ty;

    quote! {
        #[::bok_macro::repo_impl_methods(#reponame)]
        #local
    }
    .into()
}

pub(crate) fn repo_impl_methods(
    attrs: TokenStream,
    input: TokenStream,
    __source_name: TokenStream,
) -> TokenStream {
    let base = parse_macro_input!(attrs as ItemStruct);
    let local = parse_macro_input!(input as ItemImpl);
    let source_name = parse_macro_input!(__source_name as ::syn::Path);

    if let ::syn::Type::Path(self_type_path) = local.self_ty.as_ref() {
        if self_type_path.path != source_name {
            return ::syn::Error::new(
                proc_macro2::Span::call_site(),
                "#[::bok_macro::repo_impl_methods(..)]: argument and impl \
                 type differ",
            )
            .to_compile_error()
            .into();
        }
    } else {
        return ::syn::Error::new(
            proc_macro2::Span::call_site(),
            "#[::bok_macro::repo_impl_methods(..)]: argument and impl type \
             differ",
        )
        .to_compile_error()
        .into();
    }

    let local_attrs = local.attrs.iter();
    let generics = local.generics;
    let ident = &local.self_ty;
    let items = local.items.iter();

    let ::syn::Fields::Named(base_fields) = base.fields else {
        return ::syn::Error::new(
            proc_macro2::Span::call_site(),
            "#[::bok_macro::repo_impl_methods(..)]: type has unsupported \
             unnamed fields",
        )
        .to_compile_error()
        .into();
    };

    let mut fn_to_add = Vec::new();
    for f in base_fields.named.iter() {
        if !f
            .ident
            .as_ref()
            .is_some_and(|id| id.to_string().starts_with("_p_"))
        {
            continue;
        };
        let ::syn::Type::Path(f_type) = &f.ty else {
            continue;
        };
        let t_phantom = &f_type.path;
        let t_id = {
            let args = &t_phantom
                .segments
                .last()
                .expect("t_phantom no last?")
                .arguments;
            let ::syn::PathArguments::AngleBracketed(bracketed) = &args else {
                panic!("phantom without anglebracket?");
            };
            let ::syn::GenericArgument::Type(::syn::Type::Path(t)) =
                bracketed.args.first().expect("phantom bracketed, no args")
            else {
                panic!("phantom bracketed, generic not a type path");
            };
            t.path.clone()
        };
        let pkg_name = {
            let segment = t_id.segments.iter().last().unwrap();

            &segment.ident
        };
        let pkg_fn: ::syn::ImplItemFn = ::syn::parse_quote! {
            pub fn #pkg_name(&self) -> #t_id {
                #t_id::default()
            }
        };
        fn_to_add.push(pkg_fn);
    }

    let new_fn = fn_to_add.iter();

    quote! {
        #(#local_attrs)
        *
        impl #ident<#generics> {
            #(#items)
            *
            #(#new_fn)
            *
        }
    }
    .into()
}

pub(crate) fn repository(
    attrs: TokenStream,
    input: TokenStream,
) -> TokenStream {
    let local = parse_macro_input!(input as ItemStruct);
    let Fields::Named(local_fields) = local.fields else {
        use ::syn::spanned::Spanned;
        return ::syn::Error::new(
            local.fields.span(),
            "unnamed fields are not supported",
        )
        .to_compile_error()
        .into();
    };

    // do not duplicate token export or derive macro
    let local_attrs = local.attrs.iter().filter(|&x| {
        match &x.meta {
            ::syn::Meta::Path(p) => {
                // looking for:
                // #[::macro_magic::export_tokens]
                if p.segments.len() == 2
                    && p.segments[0].ident == "macro_magic"
                    && p.segments[1].ident == "export_tokens"
                {
                    false
                } else {
                    true
                }
            }
            ::syn::Meta::List(ml) => {
                // looking for:
                // #[derive(::bok::Repository, Debug)]
                if ml.path.segments.len() > 0 && ml.path.is_ident("derive") {
                    use ::syn::{punctuated::Punctuated, *};
                    if let Ok(v) = ml.parse_args_with(
                        Punctuated::<::syn::Path, Token![,]>::parse_terminated,
                    ) {
                        if v.len() == 2
                            && v[0].segments.len() == 2
                            && v[0].segments[0].ident.to_string() == "bok"
                            && v[0].segments[1].ident.to_string()
                                == "Repository"
                            && v[1].segments.len() == 1
                            && v[1].segments[0].ident.to_string() == "Debug"
                        {
                            return false;
                        }
                    }
                }
                true
            }
            _ => true,
        }
    });
    let generics = local.generics;
    let ident = local.ident;
    let vis = local.vis;

    let base = parse_macro_input!(attrs as ItemStruct);
    let Fields::Named(base_fields) = &base.fields else {
        use ::syn::spanned::Spanned;
        return ::syn::Error::new(
            base.fields.span(),
            "`#[::bok::repository(..)]`: base has unsupported unnamed fields",
        )
        .to_compile_error()
        .into();
    };
    // make sure base is a repo
    if base_fields
        .named
        .iter()
        .find(|&x| {
            x.ident
                .as_ref()
                .is_some_and(|id| id.to_string() == "_bok_repo")
        })
        .is_none()
    {
        use ::syn::spanned::Spanned;
        return ::syn::Error::new(
            base.fields.span(),
            "`#[::bok::repository(..)]` base is not a repo",
        )
        .to_compile_error()
        .into();
    }

    let mut all_fields = Vec::<::syn::Field>::with_capacity(
        local_fields.named.len() + base_fields.named.len(),
    );
    all_fields.extend(local_fields.named.iter().cloned());
    // make sure there is always `_bok_repo` marker
    if local_fields
        .named
        .iter()
        .find(|&f| {
            f.ident
                .as_ref()
                .is_some_and(|id| id.to_string() == "_bok_repo")
        })
        .is_none()
    {
        all_fields.push(::syn::parse_quote! {
            _bok_repo: ::std::marker::PhantomData<::bok::RepositoryEmpty>
        });
    }

    for b_f in base_fields.named.iter() {
        let Some(b_f_id) = &b_f.ident else { continue };
        let b_f_id_str = b_f_id.to_string();
        if all_fields
            .iter()
            .find(|&f| {
                f.ident
                    .as_ref()
                    .is_some_and(|id| id.to_string() == b_f_id_str)
            })
            .is_none()
        {
            all_fields.push(b_f.clone());
        }
    }

    quote! {
        #(#local_attrs)
        *
        #[::macro_magic::export_tokens]
        #[derive(::bok::Repository, Debug)]
        #vis struct #ident<#generics> {
            #(#all_fields),
            *
        }
    }
    .into()
}

pub(crate) fn derive_repository(input: TokenStream) -> TokenStream {
    let input = parse_macro_input!(input as DeriveInput);

    let name = input.ident.clone();

    let ::syn::Data::Struct(items) = &input.data else {
        use ::syn::spanned::Spanned;
        return ::syn::Error::new(
            input.span(),
            "#[derive(::bok::Repository)]: not called on a struct",
        )
        .to_compile_error()
        .into();
    };

    if items
        .fields
        .iter()
        .find(|&f| {
            f.ident
                .as_ref()
                .is_some_and(|id| id.to_string() == "_bok_repo")
        })
        .is_none()
    {
        use ::syn::spanned::Spanned;
        return ::syn::Error::new(
            input.span(),
            "#[derive(::bok::Repository)]: struct is not a bok repo. use \
             `#[::bok::repository(..)]` first",
        )
        .to_compile_error()
        .into();
    }
    // holds the list of all package names, snake case
    let mut all_pkgs =
        Vec::<(String, ::syn::Path)>::with_capacity(items.fields.len());

    for it in items.fields.iter() {
        let Some(id) = &it.ident else { continue };
        let id_str = id.to_string();
        if id_str.starts_with("_p_") {
            let path = {
                let ::syn::Type::Path(phantom_path) = &it.ty else {
                    continue;
                };
                let segment = phantom_path.path.segments.last().unwrap();
                let ::syn::PathArguments::AngleBracketed(bracketed) =
                    &segment.arguments
                else {
                    panic!("derive: phantom without anglebracket?");
                };
                let ::syn::GenericArgument::Type(::syn::Type::Path(t)) =
                    bracketed.args.first().expect("phantom bracketed, no args")
                else {
                    panic!(
                        "derive: phantom bracketed, generic not a type path"
                    );
                };
                t.path.clone()
            };
            let name = id_str.strip_prefix("_p_").unwrap().to_owned();
            all_pkgs.push((name, path));
        }
    }
    all_pkgs.sort_by(|a, b| a.0.cmp(&b.0));
    let pkgs_num: usize = all_pkgs.len();
    let (pkg_names, pkg_types): (Vec<String>, Vec<::syn::Path>) =
        all_pkgs.into_iter().map(|(a, b)| (a, b)).unzip();

    quote! {
        impl ::bok::Repository for #name {
            fn name(&self) -> ::bok::RepoName {
                (module_path!().to_owned() + "::" + ::std::stringify!(#name)).as_str().into()
            }
            fn path(&self) -> ::bok::Path<::bok::RepoName> {
                (module_path!().to_owned() + "::" + ::std::stringify!(#name)).as_str().into()
            }
            fn pkg_list(&self) -> &[&'static str] {
                const PKGS : [&'static str; #pkgs_num] = [
                    #(#pkg_names),
                    *
                ];
                &PKGS
            }
            fn get(&self, pkg_name: &str) -> Option<::std::boxed::Box<dyn ::bok::Pkg>> {
                match pkg_name {
                #(#pkg_names => Some(::std::boxed::Box::new(#pkg_types::default())),)
                *
                _ => None,
                }
            }
        }
    }.into()
}
struct PathList(Vec<::syn::Path>);

impl ::syn::parse::Parse for PathList {
    fn parse(input: ::syn::parse::ParseStream) -> syn::Result<Self> {
        use ::syn::punctuated::Punctuated;
        let raw =
            Punctuated::<::syn::Path, ::syn::Token![,]>::parse_terminated(
                input,
            )?;
        let mut result = Vec::with_capacity(raw.len());
        for r in raw.into_iter() {
            result.push(r)
        }
        Ok(PathList(result))
    }
}

pub(crate) fn repo_packages(
    attrs: TokenStream,
    input: TokenStream,
) -> TokenStream {
    let local = parse_macro_input!(input as ItemStruct);
    let local_attrs = local.attrs.iter();
    let generics = local.generics;
    let ident = local.ident;
    let vis = local.vis;
    use ::syn::spanned::Spanned;
    let Fields::Named(ref local_fields) = local.fields else {
        use ::syn::spanned::Spanned;
        return ::syn::Error::new(
            local.fields.span(),
            "#[repo_packages(..)]: unnamed fields are not supported",
        )
        .to_compile_error()
        .into();
    };

    // find the marker. we need it to separate things added manually
    // from things we get by extending the other repositories
    let Some(marker) = local_fields.named.iter().find(|&f| {
        f.ident
            .as_ref()
            .is_some_and(|id| id.to_string() == "_bok_repo")
    }) else {
        use ::syn::spanned::Spanned;
        return ::syn::Error::new(
            local_fields.span(),
            "#[repo_packages(..)]: struct is not a repository. Forgot \
             '#[::bok::repository(..)]` first?",
        )
        .to_compile_error()
        .into();
    };
    let fields_up_to_marker = local_fields
        .named
        .iter()
        .take_while(|&f| {
            f.ident
                .as_ref()
                .is_some_and(|id| id.to_string() != "_bok_repo")
        })
        .collect::<Vec<&::syn::Field>>();

    let mut fields_after_marker = local_fields
        .named
        .iter()
        .skip_while(|&f| {
            f.ident
                .as_ref()
                .is_some_and(|id| id.to_string() != "_bok_repo")
        })
        .skip_while(|&f| {
            f.ident
                .as_ref()
                .is_some_and(|id| id.to_string() == "_bok_repo")
        })
        .collect::<Vec<&::syn::Field>>();

    let packages = parse_macro_input!(attrs as PathList);

    // the packages added manually must not be repeated manually.
    // but they will override any other package added by
    // extending the repository
    let mut fields_added = Vec::<::syn::Field>::with_capacity(packages.0.len());
    for p in packages.0 {
        let path_ident = &p.segments.last().unwrap().ident;

        use ::convert_case::{Case, Casing};
        let pkg_id = "_p_".to_owned()
            + path_ident.to_string().to_case(Case::Snake).as_str();

        if fields_up_to_marker
            .iter()
            .find(|&f| {
                if let Some(id) = &f.ident {
                    id.to_string() == pkg_id
                } else {
                    false
                }
            })
            .is_some()
        {
            use ::syn::spanned::Spanned;
            return ::syn::Error::new(
                local_fields.span(),
                "#[repo_packages(..)]: package already present: ".to_owned()
                    + pkg_id.as_str(),
            )
            .to_compile_error()
            .into();
        }
        if fields_added
            .iter()
            .find(|&f| {
                if let Some(id) = &f.ident {
                    id.to_string() == pkg_id
                } else {
                    false
                }
            })
            .is_some()
        {
            use ::syn::spanned::Spanned;
            return ::syn::Error::new(
                local_fields.span(),
                "#[repo_packages(..)]: package added twice: ".to_owned()
                    + pkg_id.as_str(),
            )
            .to_compile_error()
            .into();
        }
        let pkg_ident =
            ::quote::format_ident!("{}", pkg_id, span = local_fields.span());

        let new_pkg: ::syn::Field = ::syn::parse_quote! {
            #pkg_ident : ::std::marker::PhantomData<#p>
        };

        fields_added.push(new_pkg);
        fields_after_marker.retain(|&f| {
            f.ident.as_ref().is_some_and(|id| id.to_string() != pkg_id)
        });
    }

    let mut all_fields =
        Vec::with_capacity(local_fields.named.len() + fields_added.len());
    all_fields.extend(fields_up_to_marker.iter());
    all_fields.extend(fields_added.iter());
    all_fields.push(marker);
    all_fields.extend(fields_after_marker.iter());

    quote! {
        #(#local_attrs)
        *
        #vis struct #ident<#generics> {
            #(#all_fields),
            *
        }
    }
    .into()
}

fn path_to_snake_case(path: &::syn::Path) -> String {
    let mut s = String::new();

    let mut is_first = true;
    for segment in path.segments.iter() {
        if !is_first {
            s += "_";
        } else {
            is_first = false
        }
        s += segment.ident.to_string().as_str();
    }

    use ::convert_case::{Case, Casing};

    s.to_case(Case::Snake)
}
